with Ada.Text_IO;

with Names;
with ModeMgr;
with UnitMgr;
with HW_Simulation;


package body AocsMgr is

   use Names;

   Units : Units_T renames UnitMgr.Units;


   procedure Sensor_Data_Processing (UnId : Unit_ID_T)
   is
      Ret_Value_from_Call : Boolean;
   begin
      Ret_Value_from_Call := HW_Simulation.Get_Sensor_Data (
         UnId, Units(UnId).CuBrId);
      if not Ret_Value_from_Call then
         Units(UnId).Br(Units(UnId).CuBrId).Error := Invalid_Data;
      end if;
   end Sensor_Data_Processing;

   procedure Actuator_Commanding (UnId : Unit_ID_T)
   is
      Ret_Value_from_Call : Boolean;
   begin
      Ret_Value_from_Call := HW_Simulation.Receive_Command (
         UnId, Units(UnId).CuBrId);
      if not Ret_Value_from_Call then
         Units(UnId).Br(Units(UnId).CuBrId).Error := Commanding_Failure;
      end if;
   end Actuator_Commanding;


   procedure Execute_Currently_Selected_Controller is
   begin

      if Control_Law.Current /= None then

         if Control_Law.Status = Preparing
            and Control_Law.Timer > 10
         then

            Control_Law.Status := Ready;

         end if;

         Control_Law.Timer := Control_Law.Timer + 1;

         if Control_Law.Current /= Virtual
            and HW_Simulation.Spurious_Error_Injection
         then
            Ada.Text_IO.Put_Line ("AOCS ERR INJECTED c="
               & Control_Law.Current'Img);

            Control_Law.Error := Attitude_Error;

         end if;

      end if;

   end Execute_Currently_Selected_Controller;


   procedure PLI_Management is
      Ret_Value_from_Call : Boolean;
   begin
      Ada.Text_IO.Put_Line ("PLI MAN br="
         & Units (PLI).CuBrId'Img
         & " t=" & Control_Law.PLI_Timer'Img
         & " r=" & PLI_Is_Ready'Img);

      Sensor_Data_Processing (PLI);

      Control_Law.PLI_Timer := Control_Law.PLI_Timer + 1;

      -- ERROR INJECTION CODE:
      --
      if Control_Law.PLI_Timer > 200 then
         Ret_Value_from_Call := PLI_Is_Ready;
         if Ret_Value_from_Call then
            Ret_Value_from_Call := HW_Simulation.Spurious_LOA_Error (
               PLI, Units (PLI).CuBrId);
            if Ret_Value_from_Call then
                Ada.Text_IO.Put_Line ("PLI ERR INJECTED br="
                       & Units (PLI).CuBrId'Img);
                Units(PLI).Br(Units(PLI).CuBrId).Error := Loss_Of_Accuracy;
            end if;
         end if;
      end if;

   end PLI_Management;


   procedure Main is
   begin

      for UnId in Unit_ID_T loop

         if Unit_Type (UnId) = Sensor
            and Units(UnId).Br(Units(UnId).CuBrId).Status = Locked
         then

            Sensor_Data_Processing (UnId);

         end if;

      end loop;

      Execute_Currently_Selected_Controller;
      -- Performs the computation of the attitude control algorithms

      for UnId in Unit_ID_T loop

         if Unit_Type (UnId) = Actuator
            and Units(UnId).Br(Units(UnId).CuBrId).Status = Locked
         then

            Actuator_Commanding (UnId);

         end if;

      end loop;

      if Units(PLI).Br(Units(PLI).CuBrId).Status = Locked
         and Control_Law.PLI_Enabled
      then

         PLI_Management;

      end if;

   end Main;


   procedure Stop_PLI_Management is
   begin

      Control_Law.PLI_Enabled := False;

   end Stop_PLI_Management;


   procedure Start_PLI_Management is
   begin

      Control_Law.PLI_Enabled := True;

      Control_Law.PLI_Timer := 0;

   end Start_PLI_Management;


   function PLI_Is_Ready return Boolean
   is
      Ret_Value : Boolean;
      Ret_Value_from_Call : Boolean;
   begin

      Ret_Value := (
         Units(PLI).Br(Units(PLI).CuBrId).Error = None
         and Units(PLI).Br(Units(PLI).CuBrId).Status = Locked
         and (Units(PLI).Br(Units(PLI).CuBrId).State = PLI_Standby
              or Units(PLI).Br(Units(PLI).CuBrId).State = PLI_Science)
         and Control_Law.PLI_Enabled);
      if Ret_Value then
         Ret_Value_from_Call := Fine_Pointing_Reached;
         Ret_Value := (Ret_Value_from_Call
            and Control_Law.PLI_Timer > 100 );
      end if;
      return Ret_Value;

   end PLI_Is_Ready;


   procedure Select_Controller (Algo : Controller_T) is
   begin

      Control_Law.Error := None;

      Control_Law.Current := Algo;

      Control_Law.Status := Idle;

      Control_Law.Timer := 0;

   end Select_Controller;


   procedure Prepare_Controller is
   begin

      Control_Law.Status := Preparing;

      Control_Law.Timer := 0;

   end Prepare_Controller;


   function Controller_Is_Ready return Boolean
   is
      Ret_Value : Boolean;
   begin

      Ret_Value := (Control_Law.Status = Ready);
      return Ret_Value;

   end Controller_Is_Ready;


   procedure Start_Controller is
   begin

      Control_Law.Status := Running;

      Control_Law.Timer := 0;

   end Start_Controller;


   function Coarse_Pointing_Reached return Boolean is
      Ret_Value : Boolean;
   begin

      Ret_Value := (
         Control_Law.Current = Coarse_Pointing_Control
         and Control_Law.Status = Running
         and Control_Law.Timer > 2000 );
      return Ret_Value;

   end Coarse_Pointing_Reached;


   function Fine_Pointing_Reached return Boolean is
      Ret_Value : Boolean;
   begin

      Ret_Value := (
         Control_Law.Current = Fine_Pointing_Control
         and Control_Law.Status = Running
         and Control_Law.Timer > 4000 );
      return Ret_Value;

   end Fine_Pointing_Reached;


end AocsMgr;
