with Ada.Text_IO;

with Names;
with AocsMgr;
with ModeMgr;
with UnitMgr;

package body FdirMgr is

   use Names;


   Units : Units_T renames UnitMgr.Units;

   Request_Mode : Target_Mode_T := None;

   procedure Handle_Attitude_Error is
   begin
      Ada.Text_IO.Put_Line ("FDIR: Handle_Attitude_Error m="
         & ModeMgr.Mode'Img
         & " t=" & ModeMgr.Trans.Target'Img);

      case ModeMgr.Mode is

         when Science =>

            Request_Mode := Preparation;

         when Preparation =>

            Request_Mode := Nominal;

         when Nominal =>

            Request_Mode := Safe;

         when others =>

            -- Emergency

            Ada.Text_IO.Put_Line ("FDIR: Request_CDMU_Reboot");
            Request_Mode := Off;

      end case;

   end Handle_Attitude_Error;


   procedure Handle_Mode_Transition_Error is
      -- Invariant: ModeMgr.Trans.Target = None
   begin
      Ada.Text_IO.Put_Line ("FDIR: Handle_Mode_Transition_Error"
         & " m=" & ModeMgr.Mode'Img
         & " t=" & ModeMgr.Trans.Target'Img);

      UnitMgr.Abort_All_Reconfigurations;

      if ModeMgr.Mode = Off
         or ModeMgr.Mode = Standby
         or ModeMgr.Trans.Target = Safe
      then
         -- Emergency

         Ada.Text_IO.Put_Line ("FDIR: Request_CDMU_Reboot");
         Request_Mode := Off;

      else

         Request_Mode := Safe;

      end if;

   end Handle_Mode_Transition_Error;


   procedure Handle_Unit_Error (UnId : Unit_ID_T)
   is
      -- Invariant: ModeMgr.Mode > Standby
      -- and Units(UnId).Br(Units(UnId).CuBrId).Status /= Free
      Ret_Value_from_Call : Boolean;
   begin
      Ada.Text_IO.Put_Line ("FDIR: Handle_Unit_Error"
         & " u=" & UnId'Img
         & " br=" & Units(UnId).CuBrId'Img
         & " err=" & Units(UnId).Br(Units(UnId).CuBrId).Error'Img
         & " c=" & Units(UnId).Br(Units(UnId).CuBrId).Error_Counter'Img);

      if UnId = PLI
         and ModeMgr.Mode = Science
         and Units(UnId).Br(Units(UnId).CuBrId).Error = Loss_Of_Accuracy
         and Units(UnId).Br(Units(UnId).CuBrId).Error_Counter < 5
      then

         Request_Mode := Preparation;

      else
          
         Ret_Value_from_Call := UnitMgr.Redundancy_Available (UnId);
         if Ret_Value_from_Call then

            UnitMgr.Start_Switch_To_Redundant (UnId);

         elsif UnId = PLI then
            -- PLI not used in Nominal Mode

            Request_Mode := Nominal;

         elsif UnitMgr.First_State (Safe, UnId) = Off then
            -- Unit not used in SafeMode

            Request_Mode := Safe;

         else
            -- Emergency: unit needed in SafeMode, and
            -- both unit branches are failed

            Ada.Text_IO.Put_Line ("FDIR: Request_CDMU_Reboot");
            Request_Mode := Off;

         end if;
      end if;

   end Handle_Unit_Error;


   procedure Main is
   begin

      if ModeMgr.Trans.Error /= None then

         Handle_Mode_Transition_Error;

      elsif AocsMgr.Control_Law.Error /= None then

         Handle_Attitude_Error;

      else

         for UnId in Unit_ID_T loop
            Ada.Text_IO.Put_Line ("FDIR unit err="
               & Units(UnId).Br(Units(UnId).CuBrId).Error'Img);
            if Units(UnId).Br(Units(UnId).CuBrId).Error /= None
               and Units(UnId).Br(Units(UnId).CuBrId).Target_State /= Off
               -- Do not handle unit which is switched off
               and (
               Units(UnId).Br(Units(UnId).CuBrId).State /= Off
               or Units(UnId).Br(Units(UnId).CuBrId).Target_State /= None )
               -- Do not reconfigure unit which is Off
               -- and not currently switched on
            then
               Units(UnId).Br(Units(UnId).CuBrId).Error_Counter :=
                  Units(UnId).Br(Units(UnId).CuBrId).Error_Counter + 1;
               Handle_Unit_Error (UnId);
               if Request_Mode /= None then
                  ModeMgr.Set_FDIR_Request (Request_Mode);
                  Request_Mode := None;
               end if;
            end if;

         end loop;

      end if;

      if Request_Mode /= None then
         ModeMgr.Set_FDIR_Request (Request_Mode);
         Request_Mode := None;
      end if;

      for UnId in Unit_ID_T loop
         Units(UnId).Br(Units(UnId).CuBrId).Error := None;
      end loop;

      ModeMgr.Trans.Error := None;

      AocsMgr.Control_Law.Error := None;

   end Main;


end FdirMgr;
