with Ada.Text_IO;

with Names;
with ModeMgr;
with UnitMgr;

package body HW_Simulation is

   use Names;

   Units : Units_T renames UnitMgr.Units;

   type HW_State_T is (
      None,
      Off,
      On,
      Extra );

   type HW_Branch_T is record
      State : HW_State_T := None;
      On_Timeout : Natural := 0;
      Extra_Timeout : Natural := 0;
   end record;

   HW_Branch : array (Unit_ID_T, Branch_ID_T) of HW_Branch_T;

   Spurious_Error_Injector : Natural := 0;

   function Spurious_Error_Injection return Boolean
   is
      -- Every call to Spurious_Error_Injection
      -- is to be modelled with a nondeterministic choice.
      Ret_Value : Boolean;
   begin
      Spurious_Error_Injector := (Spurious_Error_Injector + 1) mod 50;
      Ret_Value := (Spurious_Error_Injector = 0);
      return Ret_Value;
   end Spurious_Error_Injection;

   procedure Check_Status (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      St : Branch_Status_T;
      --
      Equal : Boolean := True)
   is
      -- PURE DEBUGGING CODE
      -- ignore this as well as the Put_Line commands
   begin
      Ada.Text_IO.Put_Line ("HWSIM check st=" & St'Img
         & " u=" & UnId'Img
         & " br=" & BrId'Img
         & " ust=" & Units(UnId).Br(BrId).Status'Img);
      if ((Equal and Units(UnId).Br(BrId).Status /= St)
          or (not Equal and Units(UnId).Br(BrId).Status = St))
         and ModeMgr.Mode /= Off
         and ModeMgr.Mode /= Standby
      then
         raise Program_Error;
      end if;
   end Check_Status;

   procedure Receive_HW_Signals (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      Is_Switch_Off : Boolean )
   is
   begin
      Ada.Text_IO.Put_Line ("HWSIM rec br="&UnId'Img & " " & BrId'Img
         & " hw=" & HW_Branch(UnId, BrId).State'Img);
      if Is_Switch_Off then
         Ada.Text_IO.Put_Line ("HWSIM rec off");
         HW_Branch(UnId, BrId).State := Off;
         HW_Branch(UnId, BrId).On_Timeout := 0;
      elsif not Spurious_Error_Injection then
         Ada.Text_IO.Put_Line ("HWSIM rec on");
         Check_Status (UnId, BrId, Reserved);
         HW_Branch(UnId, BrId).State := On;
         HW_Branch(UnId, BrId).On_Timeout := 10;
      end if;
   end Receive_HW_Signals;

   function Powered (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T )
   return Boolean
   is
      Ret_Value : Boolean;
   begin
      Ada.Text_IO.Put_Line ("HWSIM POW br="&UnId'Img & " " & BrId'Img
         & " hw=" & HW_Branch(UnId, BrId).State'Img);
      Check_Status (UnId, BrId, Reserved);
      Ret_Value := (
         HW_Branch(UnId, BrId).State = On
         and HW_Branch(UnId, BrId).On_Timeout /= 0
         and Units(UnId).Br(BrId).Timer > HW_Branch(UnId, BrId).On_Timeout);
      if Ret_Value = True then
         HW_Branch(UnId, BrId).On_Timeout := 0;
      end if;
      return Ret_Value;
   end Powered;

   procedure Receive_TC (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      TC : TC_ID_T )
   is
   begin
      Ada.Text_IO.Put_Line ("HWSIM TC br="&UnId'Img & " " & BrId'Img
         & " hw=" & HW_Branch(UnId, BrId).State'Img
         & " tc=" & TC'Img);
      Check_Status (UnId => UnId, BrId => BrId, St => Free, Equal => False);
      if not Spurious_Error_Injection then
         if HW_Branch(UnId, BrId).State = On
            and (
            (UnId = GPS and TC = GPS_Go_To_Fine_Mode)
            or (UnId = PLI and TC = PLI_Go_To_Science_Mode) )
         then
            HW_Branch(UnId, BrId).State := Extra;
            HW_Branch(UnId, BrId).Extra_Timeout := 15;
         elsif HW_Branch(UnId, BrId).State = Extra
            and (
            (UnId = GPS and TC = GPS_Go_To_Coarse_Mode)
            or (UnId = PLI and TC = PLI_Go_To_Standby_Mode) )
         then
            HW_Branch(UnId, BrId).State := On;
            HW_Branch(UnId, BrId).Extra_Timeout := 10;
         end if;
      end if;
      Ada.Text_IO.Put_Line ("HWSIM TC END br="&UnId'Img & " " & BrId'Img
         & " hw=" & HW_Branch(UnId, BrId).State'Img
         & " hwt=" & HW_Branch(UnId, BrId).Extra_Timeout'Img);
   end Receive_TC;

   function Event_Generated (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      Event : Event_ID_T )
   return Boolean
   is
      Ret_Value : Boolean;
   begin
      Ret_Value := (
         HW_Branch(UnId, BrId).Extra_Timeout /= 0
         and Units(UnId).Br(BrId).Timer
            > HW_Branch(UnId, BrId).Extra_Timeout
         and (
            (HW_Branch(UnId, BrId).State = On
             and ((UnId = GPS and Event = GPS_In_Coarse_Mode)
               or (UnId = PLI and Event = PLI_In_Standby_Mode))
            ) or
            (HW_Branch(UnId, BrId).State = Extra
             and ((UnId = GPS and Event = GPS_In_Fine_Mode)
               or (UnId = PLI and Event = PLI_In_Science_Mode))
            )
         )
      );
      if Ret_Value then
         HW_Branch(UnId, BrId).Extra_Timeout := 0;
      end if;
      Ada.Text_IO.Put_Line ("HWSIM EVENT br="&UnId'Img & " " & BrId'Img
         & " hw=" & HW_Branch(UnId, BrId).State'Img
         & " ret=" & Ret_Value'Img
         & " event=" & Event'Img
         & " hwt=" & HW_Branch(UnId, BrId).Extra_Timeout'Img
         & " t=" & Units(UnId).Br(BrId).Timer'Img);
      return Ret_Value;
   end Event_Generated;

   function Spurious_LOA_Error (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T )
   return Boolean
   is
      Ret_Value : Boolean;
   begin
      Ret_Value := (
         UnId = PLI
         and HW_Branch(UnId, BrId).State = Extra
         and HW_Branch(UnId, BrId).On_Timeout = 0
         and HW_Branch(UnId, BrId).Extra_Timeout = 0
         and Spurious_Error_Injection );
      Ada.Text_IO.Put_Line ("HWSIM PLI " & BrId'Img & " ERR=" & Ret_Value'Img);
      return Ret_Value;
   end Spurious_LOA_Error;

   function Receive_Command (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T )
   return Boolean
   is
      Ret_Value : Boolean;
   begin
      Ret_Value := (
         Unit_Type (UnId) = Actuator
         and HW_Branch(UnId, BrId).State = On
         -- Unit is switched on.
         -- Note that actuators do not have Extra state.
         and not Spurious_Error_Injection );
      Ada.Text_IO.Put_Line ("HWSIM CMD " & "u="
         & UnId'Img & " b=" & BrId'Img & " d=" & Ret_Value'Img);
      Check_Status (UnId, BrId, Locked);
      return Ret_Value;
   end Receive_Command;

   function Get_Sensor_Data (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T )
   return Boolean
   is
      Ret_Value : Boolean;
   begin
      Ret_Value := (
         Unit_Type (UnId) /= Actuator
         -- Unit is Sensor or Payload
         and (HW_Branch(UnId, BrId).State = On
            or (HW_Branch(UnId, BrId).State = Extra
                and (UnId = GPS or UnId = PLI))
         )
         -- Unit is switched on
         and not Spurious_Error_Injection );
      Ada.Text_IO.Put_Line ("HWSIM DATA " & "u="
         & UnId'Img & " b=" & BrId'Img & " d=" & Ret_Value'Img);
      Check_Status (UnId, BrId, Locked);
      return Ret_Value;
   end Get_Sensor_Data;

end HW_Simulation;
