with Ada.Characters.Handling;
with Ada.Text_IO;

with Names;
with UnitMgr.Branch_Control;

package body UnitMgr is

   use Names;

   procedure Execute_Reconf_Step (UnId : Unit_ID_T);
   -- Body declared later in this file

   procedure Main is
   begin

      for UnId in Unit_ID_T loop

         -- Continue possible unit level procedures:

         if Units(UnId).Reconf_Ongoing then

            Execute_Reconf_Step (UnId);
            -- Every unit has its own reconfiguration procedure taking
            -- care of the step condition evaluation and possible
            -- update of Units(UnId).Step. When the reconfiguration is over,
            -- branch information is updated accordingly,
            -- and Units(UnId).Reconf_Ongoing is set
            -- to False to indicate that procedure is over.

         end if;

         -- Continue possible branch level procedures:

         for BrId in Branch_ID_T loop
            Branch_Control.Possible_State_Transition_Step (UnId, BrId);
         end loop;
         Ada.Text_IO.Put ("UNITMGR u=" & UnId'Img
            & " re=" & Units(UnId).Reconf_Ongoing'Img);
         for BrId in Branch_ID_T loop
            Ada.Text_IO.Put (" "
               & Ada.Characters.Handling.To_Lower(BrId'Img)
               & "st=" & Units(UnId).Br(BrId).State'Img & " "
               & Ada.Characters.Handling.To_Lower(BrId'Img)
               & "tr=" & Units(UnId).Br(BrId).Target_State'Img & " "
               & Ada.Characters.Handling.To_Lower(BrId'Img)
               & "ss=" & Units(UnId).Br(BrId).Status'Img);
         end loop;
         Ada.Text_IO.New_Line;
      end loop;

   end Main;


   function Redundancy_Available (UnId : Unit_ID_T) return Boolean
   is
      Ret_Value : Boolean;
   begin

      Ret_Value := (
         not Units(UnId).Reconf_Ongoing
         and Units(UnId).CuBrId = A);
      return Ret_Value;

   end Redundancy_Available;


   procedure Init_Reconf_Proc (UnId : Unit_ID_T) is
   begin

      Units(UnId).Reconf_Ongoing := False;

      Units(UnId).Orig_Status   := Free;
      Units(UnId).Target_State  := Off;
      Units(UnId).Step          := 0;

   end Init_Reconf_Proc;


   procedure Abort_All_Reconfigurations is
   begin

      for UnId in Unit_ID_T loop

         Init_Reconf_Proc (UnId);

      end loop;

   end Abort_All_Reconfigurations;


   function No_Reconfiguration return Boolean is
      Ret_Value : Boolean;
   begin

      for UnId in Unit_ID_T loop

         if Units(UnId).Reconf_Ongoing then

            Ret_Value := False;
            return Ret_Value;

         end if;

      end loop;

      Ret_Value := True;
      return Ret_Value;

   end No_Reconfiguration;


   procedure Start_Switch_To_Redundant (UnId : Unit_ID_T)
   is
      -- Invariant: Units(UnId).Reconf_Ongoing = False
   begin
      Ada.Text_IO.Put_Line ("RECONF u="&UnId'Img
         & " " & Units(UnId).CuBrId'Img);

      if Units(UnId).Br(Units(UnId).CuBrId).Target_State /= Off
         -- Do not reconfigure unit which is switched off
         and (Units(UnId).Br(Units(UnId).CuBrId).State /= Off
             or Units(UnId).Br(Units(UnId).CuBrId).Target_State /= None)
         -- nor unit which is off
      then

         Units(UnId).Orig_Status :=
           Units(UnId).Br(Units(UnId).CuBrId).Status;

         if Units(UnId).Br(Units(UnId).CuBrId).Target_State /= None then
            Units(UnId).Target_State :=
              Units(UnId).Br(Units(UnId).CuBrId).Target_State;
         else
            Units(UnId).Target_State :=
              Units(UnId).Br(Units(UnId).CuBrId).State;
         end if;

         Units(UnId).Reconf_Ongoing := True;

      end if;

   end Start_Switch_To_Redundant;


   procedure Internal_Start_Trans (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      Target_State : Branch_State_T )
   is
   begin

      UnitMgr.Branch_Control.Init_Transition (UnId, BrId);

      Units(UnId).Br(BrId).Target_State := Target_State;

   end Internal_Start_Trans;


   procedure Start_Transition_Br (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      Target_State : Branch_State_T )
   is
      -- Invariants:
      -- Current branch state acceptable for Target_State
      -- other than PLI, GPS: State /= Off implies Target_State = None
      -- GPS: State = Fine_Nav and Target = Coarse_Nav implies Error = LOA
      -- GPS: to be specified
      -- PLI: to be specified
   begin

      if Target_State = Off then

         Units(UnId).Br(BrId).Error := None;
         Units(UnId).Br(BrId).Error_Counter := 0;

         if Units (UnId).Reconf_Ongoing then
            -- Request comes outside of this package,
            -- abort currently running reconfiguration

            Init_Reconf_Proc (UnId);

         end if;

         Units(UnId).Br(BrId).Status := Free;

      elsif Units(UnId).Br(BrId).Status = Free then

         for BrIdAux in Branch_ID_T loop
            Units(UnId).Br(BrIdAux).Status := Reserved;
         end loop;
         -- We reserve both branches, since the request came from
         -- ModeMgr.

      end if;

      Internal_Start_Trans (UnId, BrId, Target_State);

   end Start_Transition_Br;


   procedure Start_Transition_U (
      UnId : Unit_ID_T;
      Target_State : Branch_State_T )
   is
   begin
      Start_Transition_Br (UnId, Units(UnId).CuBrId, Target_State);
   end Start_Transition_U;

   function Trans_Done (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      State : Branch_State_T )
   return Boolean
   is
      Ret_Value : Boolean;
   begin

      Ret_Value := (
         Units(UnId).Br(BrId).State = State
         and Units(UnId).Br(BrId).Target_State = None );
      return Ret_Value;

   end Trans_Done;


   function State_by_Br_Is (
      UnId : Unit_ID_T;
      BrId : Branch_ID_T;
      State : Branch_State_T )
   return Boolean
   is
      Ret_Value : Boolean;
      Ret_Value_from_Call : Boolean;
   begin

      Ret_Value := not Units(UnId).Reconf_Ongoing;
      if Ret_Value then
         Ret_Value_from_Call := Trans_Done (UnId, BrId, State);
         Ret_Value := Ret_Value_from_Call;
      end if;
      return Ret_Value;

   end State_by_Br_Is;


   function State_by_U_Is (
      UnId : Unit_ID_T;
      State : Branch_State_T )
   return Boolean
   is
      Ret_Value : Boolean;
      Ret_Value_from_Call : Boolean;
   begin

      Ret_Value_from_Call := State_by_Br_Is (
         UnId, Units(UnId).CuBrId, State);
      Ret_Value := Ret_Value_from_Call;
      return Ret_Value;

   end State_by_U_Is;


   procedure Execute_Reconf_Step (UnId : Unit_ID_T)
   is
      -- Invariant (actually assertion at the start point):
      -- Units(UnId).Step /= 0
      -- or (Units(UnId).Br(A).State /= Off
      --     and Units(UnId).Br(B).State = Off
      --     and Units(UnId).Br(A).Status /= Free
      --     and Units(UnId).Br(B).Status = Reserved)
      Ret_Value_from_Call : Boolean;
   begin
      Ada.Text_IO.Put_Line ("RECONF STEP u="&UnId'Img
         & " " & Units(UnId).CuBrId'Img
         & " step=" & Units(UnId).Step'Img);

      case Units(UnId).Step is

         when 0 =>

            for BrId in Branch_ID_T loop
               Units(UnId).Br(BrId).Error := None;
               Units(UnId).Br(BrId).Status := Reserved;
            end loop;

            Internal_Start_Trans (UnId, A, Off);
            -- IMPORTANT Start_Transition cannot be used here

            Units(UnId).Step := Units(UnId).Step + 1;


         when 1 =>

            Ret_Value_from_Call := Trans_Done (UnId, A, Off);
            if Ret_Value_from_Call then

               Units(UnId).CuBrId := B;
               --
               -- Now A branch is off, we have to select B branch
               -- here, since FDIR monitors only "the current branch",
               -- (otherwise FDIR would not recognize failure of the
               -- branch B swiching).
               --
               -- Note that the unit status (free/reserved/lock) is
               -- changed after the correct branch state is achieved
               -- in the next step.

               Internal_Start_Trans (UnId, B, Units(UnId).Target_State);
               -- IMPORTANT Start_Transition cannot be used here

               Units(UnId).Step := Units(UnId).Step + 1;

            end if;

         when 2 =>

            Ret_Value_from_Call := Trans_Done (UnId, B, Units(UnId).Target_State);
            if Ret_Value_from_Call then

               Units(UnId).Br(B).Status := Units(UnId).Orig_Status;
               Units(UnId).Reconf_Ongoing := False;

            end if;

         when others =>

            null;

      end case;

   end Execute_Reconf_Step;

   procedure Release (UnId : Unit_ID_T) is
   begin
      for BrId in Branch_ID_T loop
         Start_Transition_Br (UnId, BrId, Off);
      end loop;
   end Release;

   procedure Release_All is
   begin
      for UnId in Unit_ID_T loop
         Release (UnId);
      end loop;
   end Release_All;

   function All_Off return Boolean is
      Ret_Value : Boolean;
      Ret_Value_from_Call : Boolean;
   begin

      for UnId in Unit_ID_T loop
         for BrId in Branch_ID_T loop
            Ret_Value_from_Call := State_by_Br_Is (UnId, BrId, Off);
            if not Ret_Value_from_Call then
               Ret_Value := False;
               return Ret_Value;
            end if;
         end loop;
      end loop;

      Ret_Value := True;
      return Ret_Value;

   end All_Off;


   procedure Start_Mode_Specific_Unit_Conf (Mode : Mode_T) is
   begin
      for UnId in Unit_ID_T loop
         if First_State (Mode, UnId) /= Off then
            if Units(UnId).Br(Units(UnId).CuBrId).State
               /= First_State (Mode, UnId)
            then
               Start_Transition_Br (UnId, Units(UnId).CuBrId,
                                    First_State (Mode, UnId));
            end if;
         else
            Release (UnId);
         end if;
      end loop;
   end Start_Mode_Specific_Unit_Conf;


   function Unit_Conf_Done (Mode : Mode_T) return Boolean is
      Ret_Value : Boolean;
      Ret_Value_from_Call : Boolean;
   begin

      for UnId in Unit_ID_T loop

         Ret_Value_from_Call := State_by_Br_Is (
            UnId, Units(UnId).CuBrId, First_State (Mode, UnId));
         if not Ret_Value_from_Call then
            Ret_Value := False;
            return Ret_Value;
         end if;

      end loop;

      Ret_Value := True;
      return Ret_Value;

   end Unit_Conf_Done;


   procedure Lock (UnId : Unit_ID_T) is
   begin

      Units(UnId).Br(Units(UnId).CuBrId).Status := Locked;

   end Lock;


   procedure Lock_All_Selected (Mode : Mode_T) is
   begin

      for UnId in Unit_ID_T loop

         if First_State (Mode, UnId) /= Off then
            -- Invariant: Units(UnId).Br(Units(UnId).CuBrId).Status = Reserved

            Lock (UnId);

         end if;

      end loop;

   end Lock_All_Selected;

end UnitMgr;
